"""
The 'worktoy.attr.field' module provides the classes used by the
comprehensive 'Field' descriptor class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations
